'use strict';

exports.__esModule = true;

var _glMatrix = require('gl-matrix');

var _euler = require('./util/euler');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var quatBuf = _glMatrix.quat.create();

var BaseTransform = function () {
  function BaseTransform() {
    var position = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : _glMatrix.vec3.create();
    var rotation = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : _glMatrix.quat.create();
    var scale = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : _glMatrix.vec3.fromValues(1, 1, 1);

    _classCallCheck(this, BaseTransform);

    this.position = position;
    this.rotation = rotation;
    this.scale = scale;

    this.localMatrix = _glMatrix.mat4.create();
    this.valid = false;
    this.ticks = 0;
  }

  BaseTransform.prototype.get = function get() {
    if (this.valid) return this.localMatrix;
    // Recalculate transform matrix
    _glMatrix.mat4.fromRotationTranslation(this.localMatrix, this.rotation, this.position);
    _glMatrix.mat4.scale(this.localMatrix, this.localMatrix, this.scale);
    this.valid = true;
    this.ticks++;
    return this.localMatrix;
  };

  BaseTransform.prototype.invalidate = function invalidate() {
    this.valid = false;
  };

  BaseTransform.prototype.identity = function identity() {
    _glMatrix.vec3.set(this.position, 0, 0, 0);
    _glMatrix.vec3.set(this.scale, 1, 1, 1);
    _glMatrix.quat.identity(this.rotation);
  };

  BaseTransform.prototype.setPos = function setPos(position) {
    _glMatrix.vec3.copy(this.position, position);
    this.invalidate();
  };

  BaseTransform.prototype.translate = function translate(position) {
    _glMatrix.vec3.add(this.position, this.position, position);
    this.invalidate();
  };

  BaseTransform.prototype.setScale = function setScale(scale) {
    _glMatrix.vec3.copy(this.scale, scale);
    this.invalidate();
  };

  BaseTransform.prototype.setAxisAngle = function setAxisAngle(axis, radian) {
    _glMatrix.quat.setAxisAngle(this.rotation, axis, radian);
    this.invalidate();
  };

  BaseTransform.prototype.rotate = function rotate(axis, radian) {
    _glMatrix.quat.setAxisAngle(quatBuf, axis, radian);
    _glMatrix.quat.multiply(this.rotation, quatBuf, this.rotation);
    this.invalidate();
  };

  BaseTransform.prototype.rotateEuler = function rotateEuler(x, y, z) {
    (0, _euler.eulerToQuat)(quatBuf, [x, y, z]);
    _glMatrix.quat.multiply(this.rotation, quatBuf, this.rotation);
    this.invalidate();
  };

  BaseTransform.prototype.rotateX = function rotateX(radian) {
    _glMatrix.quat.rotateX(this.rotation, this.rotation, radian);
    this.invalidate();
  };

  BaseTransform.prototype.rotateY = function rotateY(radian) {
    _glMatrix.quat.rotateY(this.rotation, this.rotation, radian);
    this.invalidate();
  };

  BaseTransform.prototype.rotateZ = function rotateZ(radian) {
    _glMatrix.quat.rotateZ(this.rotation, this.rotation, radian);
    this.invalidate();
  };

  BaseTransform.prototype.lookAt = function lookAt(front, up) {
    var zFront = _glMatrix.vec3.create();
    _glMatrix.vec3.normalize(zFront, front);
    var right = _glMatrix.vec3.create();
    _glMatrix.vec3.cross(right, up, zFront);
    _glMatrix.vec3.normalize(right, right);
    if (_glMatrix.vec3.length(right) < 0.0001) {
      right = [1, 0, 0];
    }
    var yUp = _glMatrix.vec3.create();
    _glMatrix.vec3.cross(yUp, zFront, right);
    var mat = [right[0], right[1], right[2], yUp[0], yUp[1], yUp[2], zFront[0], zFront[1], zFront[2]];
    _glMatrix.quat.fromMat3(this.rotation, mat);
    this.invalidate();
  };

  BaseTransform.prototype.getEuler = function getEuler(out) {
    return (0, _euler.quatToEuler)(out, this.rotation);
  };

  return BaseTransform;
}();

exports.default = BaseTransform;