'use strict';

exports.__esModule = true;

var _glMatrix = require('gl-matrix');

var _camera = require('../camera');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function easeInOutQuad(t) {
  t *= 2;
  if (t < 1) return t * t / 2;
  t--;
  return (t * (t - 2) - 1) / -2;
}

var BlenderCameraController = function () {
  function BlenderCameraController(node, keyNode, camera) {
    _classCallCheck(this, BlenderCameraController);

    // Blender-like control mode
    this.center = _glMatrix.vec3.create();
    this.radius = 6;

    this.mouseHeld = false;
    this.mouseX = 0;
    this.mouseY = 0;
    this.rotateDir = 0;

    this.slerpStart = _glMatrix.quat.create();
    this.slerpEnd = _glMatrix.quat.create();
    this.slerpCounter = -1;

    this.lerpStart = _glMatrix.vec3.create();
    this.lerpEnd = _glMatrix.vec3.create();
    this.lerpCounter = -1;

    // false - Blender-like control
    // true - FPS-like control
    this.mode = false;
    //
    this.perspective = true;

    this.hasChanged = true;
    this.camera = camera;

    this.node = node;
    this.keyNode = keyNode;

    this.registerEvents();
  }

  BlenderCameraController.prototype.registerEvents = function registerEvents() {
    var _this = this;

    this.node.addEventListener('mousemove', function (e) {
      if (!_this.mouseHeld) return;
      var offsetX = e.clientX - _this.mouseX;
      var offsetY = e.clientY - _this.mouseY;
      _this.mouseX = e.clientX;
      _this.mouseY = e.clientY;
      var transform = _this.camera.transform;
      if (e.shiftKey) {
        // Do translation instead - we'd need two vectors to make translation
        // relative to the camera rotation
        var vecX = _glMatrix.vec3.create();
        var vecY = _glMatrix.vec3.create();
        _glMatrix.vec3.transformQuat(vecX, [-offsetX * _this.radius / 600, 0, 0], transform.rotation);
        _glMatrix.vec3.transformQuat(vecY, [0, offsetY * _this.radius / 600, 0], transform.rotation);
        _glMatrix.vec3.add(_this.center, _this.center, vecX);
        _glMatrix.vec3.add(_this.center, _this.center, vecY);
        _this.hasChanged = true;
        return;
      }
      // rotation....
      var rot = _glMatrix.quat.create();
      _glMatrix.quat.rotateY(rot, rot, Math.PI / 180 * -offsetX * _this.rotateDir / 4);
      _glMatrix.quat.multiply(transform.rotation, rot, transform.rotation);
      _glMatrix.quat.rotateX(transform.rotation, transform.rotation, Math.PI / 180 * -offsetY / 4);
      _this.hasChanged = true;
    });
    this.node.addEventListener('contextmenu', function (e) {
      e.preventDefault();
    });
    this.node.addEventListener('touchstart', function (e) {
      e.preventDefault();
      _this.mouseHeld = true;
      // Determine if we should go clockwise or anticlockwise.
      var upLocal = _glMatrix.vec3.create();
      var up = _glMatrix.vec3.fromValues(0, 1, 0);
      _glMatrix.vec3.transformQuat(upLocal, [0, 1, 0], _this.camera.transform.rotation);
      var upDot = _glMatrix.vec3.dot(up, upLocal);
      _this.rotateDir = upDot >= 0 ? 1 : -1;
      // Set position
      _this.mouseX = e.changedTouches[0].pageX;
      _this.mouseY = e.changedTouches[0].pageY;
    }, false);
    this.node.addEventListener('touchmove', function (e) {
      if (!_this.mouseHeld) return;
      var offsetX = e.changedTouches[0].pageX - _this.mouseX;
      var offsetY = e.changedTouches[0].pageY - _this.mouseY;
      _this.mouseX = e.changedTouches[0].pageX;
      _this.mouseY = e.changedTouches[0].pageY;
      var transform = _this.camera.transform;
      // rotation....
      var rot = _glMatrix.quat.create();
      _glMatrix.quat.rotateY(rot, rot, Math.PI / 180 * -offsetX * _this.rotateDir / 4);
      _glMatrix.quat.multiply(transform.rotation, rot, transform.rotation);
      _glMatrix.quat.rotateX(transform.rotation, transform.rotation, Math.PI / 180 * -offsetY / 4);
      _this.hasChanged = true;
    });
    this.node.addEventListener('touchend', function (e) {
      e.preventDefault();
      _this.mouseHeld = false;
    }, false);
    this.node.addEventListener('touchcancel', function (e) {
      e.preventDefault();
      _this.mouseHeld = false;
    }, false);
    this.node.addEventListener('mousedown', function (e) {
      if (e.button === 0) return;
      _this.mouseHeld = true;
      // Determine if we should go clockwise or anticlockwise.
      var upLocal = _glMatrix.vec3.create();
      var up = _glMatrix.vec3.fromValues(0, 1, 0);
      _glMatrix.vec3.transformQuat(upLocal, [0, 1, 0], _this.camera.transform.rotation);
      var upDot = _glMatrix.vec3.dot(up, upLocal);
      _this.rotateDir = upDot >= 0 ? 1 : -1;
      // Set position
      _this.mouseX = e.clientX;
      _this.mouseY = e.clientY;
      e.preventDefault();
    });
    this.node.addEventListener('mouseup', function (e) {
      if (e.button === 0) return;
      _this.mouseHeld = false;
      e.preventDefault();
    });
    this.keyNode.addEventListener('keydown', function (e) {
      if (e.shiftKey) return;
      var camera = _this.camera;

      if (e.keyCode === 32) {
        _glMatrix.vec3.copy(_this.lerpStart, _this.center);
        _glMatrix.vec3.set(_this.lerpEnd, 0, 0, 0);
        _this.lerpCounter = 0;
      }
      // Persp - Ortho swap
      if (e.keyCode === 101 || e.keyCode === 53) {
        _this.perspective = !_this.perspective;
        if (_this.perspective) {
          camera.projection = (0, _camera.perspective)(Math.PI / 180 * 70, 0.3, 1000);
        }
        _this.hasChanged = true;
      }
      // Front
      if (e.keyCode === 97 || e.keyCode === 49) {
        _glMatrix.quat.copy(_this.slerpStart, camera.transform.rotation);
        _glMatrix.quat.identity(_this.slerpEnd);
        if (e.ctrlKey) {
          _glMatrix.quat.rotateY(_this.slerpEnd, _this.slerpEnd, Math.PI);
        }
        _this.slerpCounter = 0;
      }
      // Right
      if (e.keyCode === 99 || e.keyCode === 51) {
        _glMatrix.quat.copy(_this.slerpStart, camera.transform.rotation);
        _glMatrix.quat.identity(_this.slerpEnd);
        _glMatrix.quat.rotateY(_this.slerpEnd, _this.slerpEnd, Math.PI / 2);
        if (e.ctrlKey) {
          _glMatrix.quat.rotateY(_this.slerpEnd, _this.slerpEnd, -Math.PI);
        }
        _this.slerpCounter = 0;
      }
      // Top
      if (e.keyCode === 103 || e.keyCode === 55) {
        _glMatrix.quat.copy(_this.slerpStart, camera.transform.rotation);
        _glMatrix.quat.identity(_this.slerpEnd);
        _glMatrix.quat.rotateX(_this.slerpEnd, _this.slerpEnd, -Math.PI / 2);
        if (e.ctrlKey) {
          _glMatrix.quat.rotateX(_this.slerpEnd, _this.slerpEnd, Math.PI);
        }
        _this.slerpCounter = 0;
      }
    });
    this.node.addEventListener('wheel', function (e) {
      var diff = e.deltaY / 50;
      if (e.deltaMode === 0) diff /= 12;
      if (e.shiftKey) {
        var vecY = _glMatrix.vec3.create();
        _glMatrix.vec3.transformQuat(vecY, [0, -diff * _this.radius, 0], _this.rotation);
        _glMatrix.vec3.add(_this.center, _this.center, vecY);
        _this.hasChanged = true;
        e.preventDefault();
        return;
      } else if (e.ctrlKey) {
        var vecX = _glMatrix.vec3.create();
        _glMatrix.vec3.transformQuat(vecX, [diff * _this.radius, 0, 0], _this.rotation);
        _glMatrix.vec3.add(_this.center, _this.center, vecX);
        _this.hasChanged = true;
        e.preventDefault();
        return;
      }
      _this.radius += _this.radius * diff;
      _this.hasChanged = true;
      e.preventDefault();
    });
  };

  BlenderCameraController.prototype.update = function update(delta) {
    var transform = this.camera.transform;
    if (this.lerpCounter !== -1) {
      this.lerpCounter = Math.min(1, this.lerpCounter + delta * 4);
      _glMatrix.vec3.lerp(this.center, this.lerpStart, this.lerpEnd, easeInOutQuad(this.lerpCounter));
      this.hasChanged = true;
      if (this.lerpCounter >= 1) this.lerpCounter = -1;
    }
    if (this.slerpCounter !== -1) {
      this.slerpCounter = Math.min(1, this.slerpCounter + delta * 4);
      _glMatrix.quat.slerp(transform.rotation, this.slerpStart, this.slerpEnd, easeInOutQuad(this.slerpCounter));
      this.hasChanged = true;
      if (this.slerpCounter >= 1) this.slerpCounter = -1;
    }
    if (this.hasChanged) {
      if (!this.perspective) {
        this.camera.projection = (0, _camera.orthogonal)(this.radius / 2, 0.1, 1000);
      }
      _glMatrix.vec3.transformQuat(transform.position, [0, 0, this.radius], transform.rotation);
      _glMatrix.vec3.add(transform.position, transform.position, this.center);
      transform.invalidate();
      this.hasChanged = false;
    }
  };

  return BlenderCameraController;
}();

exports.default = BlenderCameraController;