'use strict';

exports.__esModule = true;
exports.default = calcTangents;

var _parseAttributes = require('../../util/parseAttributes');

var _parseIndices = require('../../util/parseIndices');

var _parseIndices2 = _interopRequireDefault(_parseIndices);

var _createIndicesArray = require('../../util/createIndicesArray');

var _createIndicesArray2 = _interopRequireDefault(_createIndicesArray);

var _glMatrix = require('gl-matrix');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function calcTangents(input) {
  var positions = (0, _parseAttributes.parseAttribute)(input.attributes.aPosition);
  var indices = (0, _parseIndices2.default)(input.indices.aPosition);
  var texCoords = (0, _parseAttributes.parseAttribute)(input.attributes.aTexCoord);
  var texIndices = (0, _parseIndices2.default)(input.indices.aTexCoord);
  if (positions == null || indices == null) {
    throw new Error('Position attribute and indices must be specified');
  }
  if (texCoords == null || texIndices == null) {
    throw new Error('Position attribute and indices must be specified');
  }
  // One vec4 per one face. so indices / 3 * 4.
  var tangents = new Float32Array(indices.length / 3 * 4);
  var tangentIndices = (0, _createIndicesArray2.default)(indices.length / 3, indices.length);
  var tangentId = 0;
  for (var faceId = 0; faceId < indices.length; faceId += 3) {
    var vertexId1 = indices[faceId];
    var vertexId2 = indices[faceId + 1];
    var vertexId3 = indices[faceId + 2];
    var texId1 = texIndices[faceId];
    var texId2 = texIndices[faceId + 1];
    var texId3 = texIndices[faceId + 2];
    // Calculate tangent vector.
    var origin = positions.data.subarray(vertexId1 * 3, vertexId1 * 3 + 3);
    var p1 = _glMatrix.vec3.create();
    _glMatrix.vec3.subtract(p1, positions.data.subarray(vertexId2 * 3, vertexId2 * 3 + 3), origin);
    var p2 = _glMatrix.vec3.create();
    _glMatrix.vec3.subtract(p2, positions.data.subarray(vertexId3 * 3, vertexId3 * 3 + 3), origin);
    var texOrigin = texCoords.data.subarray(texId1 * 2, texId1 * 2 + 2);
    var texP1 = _glMatrix.vec2.create();
    _glMatrix.vec2.subtract(texP1, texCoords.data.subarray(texId2 * 2, texId2 * 2 + 2), texOrigin);
    var texP2 = _glMatrix.vec2.create();
    _glMatrix.vec2.subtract(texP2, texCoords.data.subarray(texId3 * 2, texId3 * 2 + 2), texOrigin);
    // Honestly I don't know what this does.
    var f = 1 / (texP1[0] * texP2[1] - texP2[0] * texP1[1]);
    var tangent = new Float32Array(4);
    tangent[0] = f * (texP2[1] * p1[0] - texP1[1] * p2[0]);
    tangent[1] = f * (texP2[1] * p1[1] - texP1[1] * p2[1]);
    tangent[2] = f * (texP2[1] * p1[2] - texP1[1] * p2[2]);
    // Calculate bi-tangent. To save vertex array, it can be calculated in
    // vertex shader; however we have to specify the cross order to get right
    // result. This can be done by using a modifier... I think.
    // To calculate modifier, we have to calculate dot product with
    // bi-tangent from vertex shader and bi-tangent we calculated.
    var normal = _glMatrix.vec3.create();
    _glMatrix.vec3.cross(normal, p1, p2);
    _glMatrix.vec3.normalize(normal, normal);
    var leftBitangent = _glMatrix.vec3.create();
    _glMatrix.vec3.cross(leftBitangent, tangent, normal);
    // Then calculate bi-tangent with texture coords.
    var bitangent = _glMatrix.vec3.create();
    bitangent[0] = f * (texP2[0] * p1[0] - texP1[0] * p2[0]);
    bitangent[1] = f * (texP2[0] * p1[1] - texP1[0] * p2[1]);
    bitangent[2] = f * (texP2[0] * p1[2] - texP1[0] * p2[2]);
    var modifier = _glMatrix.vec3.dot(bitangent, leftBitangent);
    tangent[3] = modifier < 0 ? -1 : 1;
    // Done. Store them in tangents buffer, and set indices to it.
    tangents.set(tangent, tangentId * 4);
    tangentIndices[faceId] = tangentId;
    tangentIndices[faceId + 1] = tangentId;
    tangentIndices[faceId + 2] = tangentId;
    tangentId++;
  }
  return Object.assign({}, input, {
    attributes: Object.assign({}, input.attributes, {
      aPosition: positions, aTexCoord: texCoords,
      aTangent: { axis: 4, data: tangents }
    }),
    indices: Object.assign({}, input.indices, {
      aPosition: indices, aTexCoord: texIndices, aTangent: tangentIndices
    })
  });
}