'use strict';

exports.__esModule = true;
exports.default = combine;

var _parseAttributes = require('../util/parseAttributes');

var _parseAttributes2 = _interopRequireDefault(_parseAttributes);

var _parseIndices = require('../util/parseIndices');

var _parseIndices2 = _interopRequireDefault(_parseIndices);

var _createIndicesArray = require('../util/createIndicesArray');

var _createIndicesArray2 = _interopRequireDefault(_createIndicesArray);

var _unwrapInstanced = require('./unwrapInstanced');

var _unwrapInstanced2 = _interopRequireDefault(_unwrapInstanced);

var _geometry = require('../renderer/geometry');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function flatten(arr) {
  return arr.reduce(function (prev, t) {
    return prev.concat(Array.isArray(t) ? flatten(t) : t);
  }, []);
}

function combine(input) {
  if (!Array.isArray(input) || input.length === 0) {
    throw new Error('Input must be an array with more than one geometry');
  }
  // Flatten the array, then parse the attributes and indices first
  var geometries = flatten(input).map(function (geometry) {
    // If instanced, we must unwrap it to combine it to single draw call.
    // But this behavior is kinda weird though.
    if (geometry.instanced) return (0, _unwrapInstanced2.default)(geometry);
    return {
      attributes: (0, _parseAttributes2.default)(geometry.attributes),
      indices: (0, _parseIndices2.default)(geometry.indices),
      mode: geometry.mode
    };
  });
  // Merge geometries by their modes
  // TODO If the mode is not one of POINTS, LINES, TRIANGLES, they must be
  // unwrapped to one of these to be merged.
  // First, probe the complete list of modes
  var modes = geometries.reduce(function (modes, geometry) {
    var geomMode = geometry.mode == null ? _geometry.TRIANGLES : geometry.mode;
    if (modes.indexOf(geomMode) !== -1) return modes;
    return modes.concat(geomMode);
  }, []);
  if (modes.length > 1) {
    // More than one modes exist, Run combine for each modes.
    var modeGeoms = modes.map(function (mode) {
      return geometries.filter(function (geometry) {
        return (geometry.mode == null ? _geometry.TRIANGLES : geometry.mode) === mode;
      });
    });
    // Done
    return modeGeoms.map(combine);
  }
  var mode = modes[0];
  // First, calculate vertices count per geometry, while creating attribute
  // information (side-effect).
  var attributes = {};
  var vertexCounts = geometries.map(function (geometry) {
    var vertexCount = -1;
    for (var i in geometry.attributes) {
      var attribute = geometry.attributes[i];
      if (vertexCount === -1) {
        vertexCount = attribute.data.length / attribute.axis;
      } else if (vertexCount !== attribute.data.length / attribute.axis) {
        throw new Error('Vertex count mismatch');
      }
      // Create attribute data
      if (attributes[i] == null) {
        attributes[i] = {
          axis: attribute.axis
        };
      } else {
        attributes[i].axis = Math.max(attributes[i].axis, attribute.axis);
      }
    }
    if (vertexCount === -1) {
      throw new Error('Geometry should have at least one attribute');
    }
    return vertexCount;
  });
  var vertexCount = vertexCounts.reduce(function (a, b) {
    return a + b;
  });
  // Then, calculate total indices count. We have to create 'linear' indices
  // if a geometry doesn't have indices.
  var useIndices = geometries.some(function (a) {
    return a.indices;
  });
  var indicesCount = geometries.reduce(function (count, geometry, index) {
    if (geometry.indices == null) {
      return count + vertexCounts[index];
    }
    return count + geometry.indices.length;
  }, 0);
  // Pre-calculation is complete, populate actual data (per attribute)

  var _loop = function _loop(key) {
    // TODO Support other than Float32Array
    var attribute = attributes[key];
    attribute.data = new Float32Array(vertexCount * attribute.axis);
    // Actually populate data
    var offset = 0;
    geometries.forEach(function (geometry, i) {
      if (geometry.attributes[key]) {
        var geomAttribute = geometry.attributes[key];
        // Copy the whole data upon it (if axis matches)
        if (attribute.axis === geomAttribute.axis) {
          attribute.data.set(geomAttribute.data, offset);
        } else {
          // Otherwise, we have to copy it by one by one
          for (var j = 0; j < vertexCounts[i]; ++j) {
            var subset = geomAttribute.data.subarray(j, j + geomAttribute.axis);
            attribute.data.set(subset, offset + j * attribute.axis);
          }
        }
      }
      offset += vertexCounts[i] * attribute.axis;
    });
  };

  for (var key in attributes) {
    _loop(key);
  }
  var indices = void 0;
  // Now, populate the indices
  if (useIndices) {
    indices = (0, _createIndicesArray2.default)(vertexCount, indicesCount);
    var indicesOffset = 0;
    var verticesOffset = 0;
    geometries.forEach(function (geometry, i) {
      if (geometry.indices) {
        for (var j = 0; j < geometry.indices.length; ++j) {
          indices[indicesOffset + j] = geometry.indices[j] + verticesOffset;
        }
      } else {
        for (var _j = 0; _j < vertexCounts[i]; ++_j) {
          indices[indicesOffset + _j] = _j + verticesOffset;
        }
      }
      indicesOffset += geometry.indices ? geometry.indices.length : vertexCounts[i];
      verticesOffset += vertexCounts[i];
    });
  }
  return {
    mode: mode, attributes: attributes, indices: indices
  };
}