'use strict';

exports.__esModule = true;
exports.default = transform;

var _parseAttributes = require('../util/parseAttributes');

var _parseAttributes2 = _interopRequireDefault(_parseAttributes);

var _getVerticesCount = require('../util/getVerticesCount');

var _getVerticesCount2 = _interopRequireDefault(_getVerticesCount);

var _glMatrix = require('gl-matrix');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function getTransformer(axis, transform) {
  // If function, just use it
  if (typeof transform === 'function') return transform;
  // Otherwise, try to find matching transformer
  if (Array.isArray(transform) || transform instanceof Float32Array) {
    if (axis === 2 && transform.length === 4) {
      return function (v) {
        return _glMatrix.vec2.transformMat2(v, v, transform);
      };
    } else if (axis === 2 && transform.length === 6) {
      return function (v) {
        return _glMatrix.vec2.transformMat2d(v, v, transform);
      };
    } else if (axis === 2 && transform.length === 9) {
      return function (v) {
        return _glMatrix.vec2.transformMat3(v, v, transform);
      };
    } else if (axis === 3 && transform.length === 9) {
      return function (v) {
        return _glMatrix.vec3.transformMat3(v, v, transform);
      };
    } else if (axis === 3 && transform.length === 16) {
      return function (v) {
        return _glMatrix.vec3.transformMat4(v, v, transform);
      };
    } else if (axis === 4 && transform.length === 16) {
      return function (v) {
        return _glMatrix.vec4.transformMat4(v, v, transform);
      };
    } else if (axis === transform.length) {
      return function (v) {
        return v.set(transform);
      };
    } else {
      throw new Error('Unknown axis / transformer pair');
    }
  } else {
    throw new Error('Unknown transformer type');
  }
}

function transform(input, transforms) {
  if (Array.isArray(input)) return input.map(function (v) {
    return transform(v, transforms);
  });
  // Create exact copy of the input, with some data modified.
  var attributes = (0, _parseAttributes2.default)(input.attributes);
  var verticesCount = -1;
  for (var key in transforms) {
    var attribute = void 0;
    if (attributes[key] == null) {
      if (verticesCount === -1) {
        verticesCount = (0, _getVerticesCount2.default)(input.attributes);
      }
      // Special case - if attribute doesn't exists and transformer's size
      // equals or is smaller than 4, create one.
      // TODO This can cause a confliction with 2x2 transform matrix
      if ((Array.isArray(transforms[key]) || transforms[key] instanceof Float32Array) && transforms[key].length <= 4) {
        attribute = attributes[key] = {
          axis: transforms[key].length,
          data: new Float32Array(verticesCount * transforms[key].length)
        };
      } else {
        continue;
      }
    } else {
      attribute = attributes[key];
    }
    // Since geometry functions should be immutable, we must create a clone
    // to work with
    var data = attribute.data.slice();
    var axis = attribute.axis;
    var _transform = getTransformer(axis, transforms[key]);
    // Infinite loop protection, though it can't protect NaN
    if (axis <= 0) throw new Error('Axis must be larger than 0');
    for (var i = 0; i < attribute.data.length; i += axis) {
      var value = data.subarray(i, i + axis);
      // Now, change the value (We can do it safely because we're working with
      // the clone)
      _transform(value, i);
    }
    // Tada! Overwrite the attributes.
    attributes[key] = { axis: axis, data: data };
  }
  return Object.assign({}, input, { attributes: attributes });
}