"use strict";

exports.__esModule = true;
exports.default = uvSphere;
function uvSphere(segments, rings) {
  // A UV Sphere consists of many regular N-polygons, which serves as
  // a 'ring'. Its radius gracefully goes small, and eventually reach 0
  // at the top. A sphere is defined as x^2 + y^2 + z^2 = 1. Which converts
  // to 1 - z^2 = x^2 + y^2. We can derive that a ring's radius is
  // sqrt(1 - z^2) from that equation.
  // The sphere has ring * (segment - 2) + 2 vertices. Since radius 0 means
  // a single point, we can remove 2 rings from top and bottom.
  // However, OpenGL requires another vertices for different normals, etc,
  // we must use different vertices for each faces.
  // A UV sphere can be seperated to 3 parts - Roof, Pillar, Floor.
  // Roof consists of N triangles (N = segments), so it'd be trivial -
  // it needs 3 * N vertices.
  // Pillar, however, uses quads. However, we can share the vertices if
  // it has same normal map, texture coords, etc.
  // Pillar consists of N * (M - 3) quads (M = rings), so it uses
  // 4 * N * (M - 3) vertices.
  // Floor is exactly same as the roof, it needs 3 * N vertices.
  // Adding all, it needs N * (6 + 4 * (M - 3)) vertices total, and
  // N * (2 + 2 * (M - 3)) triangles total.
  var verticesTotal = segments * (6 + 4 * (rings - 3));
  var facesTotal = segments * (2 + 2 * (rings - 3));
  var texCoords = new Float32Array(verticesTotal * 2);
  var vertices = new Float32Array(verticesTotal * 3);
  var indices = new Uint16Array(facesTotal * 3);
  // Roof
  for (var i = 0; i < segments; ++i) {
    var theta = 1 / (rings - 1) * Math.PI;
    var y = Math.cos(theta);
    var radius = Math.sin(theta);
    var angle = i / segments * Math.PI * 2;
    var angleNext = (i + 1) / segments * Math.PI * 2;
    // Roof left
    vertices[i * 9] = Math.cos(angle) * radius;
    vertices[i * 9 + 1] = y;
    vertices[i * 9 + 2] = Math.sin(angle) * radius;
    // Roof top
    vertices[i * 9 + 3] = 0;
    vertices[i * 9 + 4] = 1;
    vertices[i * 9 + 5] = 0;
    // Roof right
    vertices[i * 9 + 6] = Math.cos(angleNext) * radius;
    vertices[i * 9 + 7] = y;
    vertices[i * 9 + 8] = Math.sin(angleNext) * radius;
    // Connect it
    indices[i * 3] = i * 3;
    indices[i * 3 + 1] = i * 3 + 1;
    indices[i * 3 + 2] = i * 3 + 2;
    // Texture coords
    texCoords[i * 6] = 1 - i / segments;
    texCoords[i * 6 + 1] = 1 - 1 / (rings - 1);
    texCoords[i * 6 + 2] = 1 - (i + 0.5) / segments;
    texCoords[i * 6 + 3] = 1;
    texCoords[i * 6 + 4] = 1 - (i + 1) / segments;
    texCoords[i * 6 + 5] = 1 - 1 / (rings - 1);
  }
  // Pillar
  for (var j = 0; j < rings - 3; ++j) {
    var _theta = (j + 1) / (rings - 1) * Math.PI;
    var _y = Math.cos(_theta);
    var _radius = Math.sin(_theta);
    var thetaNext = (j + 2) / (rings - 1) * Math.PI;
    var yNext = Math.cos(thetaNext);
    var radiusNext = Math.sin(thetaNext);
    for (var _i = 0; _i < segments; ++_i) {
      var startPos = segments * (3 + 4 * j) + _i * 4;
      var startIndices = segments * (3 + 6 * j) + _i * 6;
      var _angle = _i / segments * Math.PI * 2;
      var _angleNext = (_i + 1) / segments * Math.PI * 2;
      // Roof left top
      vertices[startPos * 3] = Math.cos(_angle) * _radius;
      vertices[startPos * 3 + 1] = _y;
      vertices[startPos * 3 + 2] = Math.sin(_angle) * _radius;
      // Roof right top
      vertices[startPos * 3 + 3] = Math.cos(_angleNext) * _radius;
      vertices[startPos * 3 + 4] = _y;
      vertices[startPos * 3 + 5] = Math.sin(_angleNext) * _radius;
      // Roof right bottom
      vertices[startPos * 3 + 6] = Math.cos(_angleNext) * radiusNext;
      vertices[startPos * 3 + 7] = yNext;
      vertices[startPos * 3 + 8] = Math.sin(_angleNext) * radiusNext;
      // Roof left bottom
      vertices[startPos * 3 + 9] = Math.cos(_angle) * radiusNext;
      vertices[startPos * 3 + 10] = yNext;
      vertices[startPos * 3 + 11] = Math.sin(_angle) * radiusNext;
      // Connect it, kinda gross though
      indices[startIndices] = startPos + 2;
      indices[startIndices + 1] = startPos;
      indices[startIndices + 2] = startPos + 1;
      indices[startIndices + 3] = startPos + 3;
      indices[startIndices + 4] = startPos;
      indices[startIndices + 5] = startPos + 2;
      // Texture coords mapping
      texCoords[startPos * 2] = 1 - _i / segments;
      texCoords[startPos * 2 + 1] = 1 - (j + 1) / (rings - 1);
      texCoords[startPos * 2 + 2] = 1 - (_i + 1) / segments;
      texCoords[startPos * 2 + 3] = 1 - (j + 1) / (rings - 1);
      texCoords[startPos * 2 + 4] = 1 - (_i + 1) / segments;
      texCoords[startPos * 2 + 5] = 1 - (j + 2) / (rings - 1);
      texCoords[startPos * 2 + 6] = 1 - _i / segments;
      texCoords[startPos * 2 + 7] = 1 - (j + 2) / (rings - 1);
    }
  }
  // Floor
  for (var _i2 = 0; _i2 < segments; ++_i2) {
    var _startPos = segments * (3 + 4 * (rings - 3)) + _i2 * 3;
    var _startIndices = segments * (3 + 6 * (rings - 3)) + _i2 * 3;
    var _theta2 = (rings - 2) / (rings - 1) * Math.PI;
    var _y2 = Math.cos(_theta2);
    var _radius2 = Math.sin(_theta2);
    var _angle2 = _i2 / segments * Math.PI * 2;
    var _angleNext2 = (_i2 + 1) / segments * Math.PI * 2;
    // Floor left
    vertices[_startPos * 3] = Math.cos(_angle2) * _radius2;
    vertices[_startPos * 3 + 1] = _y2;
    vertices[_startPos * 3 + 2] = Math.sin(_angle2) * _radius2;
    // Floor top
    vertices[_startPos * 3 + 3] = 0;
    vertices[_startPos * 3 + 4] = -1;
    vertices[_startPos * 3 + 5] = 0;
    // Floor right
    vertices[_startPos * 3 + 6] = Math.cos(_angleNext2) * _radius2;
    vertices[_startPos * 3 + 7] = _y2;
    vertices[_startPos * 3 + 8] = Math.sin(_angleNext2) * _radius2;
    // Connect it
    indices[_startIndices] = _startPos + 1;
    indices[_startIndices + 1] = _startPos;
    indices[_startIndices + 2] = _startPos + 2;
    // Texture coords
    texCoords[_startPos * 2] = 1 - _i2 / segments;
    texCoords[_startPos * 2 + 1] = 1 - (rings - 2) / (rings - 1);
    texCoords[_startPos * 2 + 2] = 1 - (_i2 + 0.5) / segments;
    texCoords[_startPos * 2 + 3] = 0;
    texCoords[_startPos * 2 + 4] = 1 - (_i2 + 1) / segments;
    texCoords[_startPos * 2 + 5] = 1 - (rings - 2) / (rings - 1);
  }
  return {
    attributes: {
      aPosition: { axis: 3, data: vertices },
      aTexCoord: { axis: 2, data: texCoords },
      // Instead of calculating normals by calculating indices, we can
      // just put vertices data to the normals to calculate smooth
      // normals. (This only applies to sphere though)
      aNormal: { axis: 3, data: vertices }
    },
    indices: indices
  };
}