'use strict';

exports.__esModule = true;
exports.default = loadOBJ;

var _calcTangents = require('../geom/channel/calcTangents');

var _calcTangents2 = _interopRequireDefault(_calcTangents);

var _calcNormals = require('../geom/channel/calcNormals');

var _calcNormals2 = _interopRequireDefault(_calcNormals);

var _calcSmoothNormals = require('../geom/channel/calcSmoothNormals');

var _calcSmoothNormals2 = _interopRequireDefault(_calcSmoothNormals);

var _parseIndices = require('../util/parseIndices');

var _parseIndices2 = _interopRequireDefault(_parseIndices);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Loads OBJ file to Geometry object. Currently does not support materials
// and etc.
function loadOBJ(data) {
  var separate = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
  var noObject = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

  // Parser machine state
  var objName = 'OBJ_' + (Math.random() * 1000 | 0);
  var objMaterial = null;
  var vertices = [];
  var normals = [];
  var texCoords = [];
  var vertexIndices = [];
  var normalIndices = [];
  var texCoordIndices = [];
  var normalSpecified = false;
  var normalSmooth = false;

  var attributes = void 0;
  var polyList = [];
  var geometries = noObject ? [] : {};

  function addFaceIndex(point) {
    vertexIndices.push(point.vertex);
    texCoordIndices.push(point.texCoord);
    normalIndices.push(point.normal);
  }

  function finalizePolyList() {
    if (vertexIndices.length === 0) return;
    var geometry = {
      metadata: {
        material: objMaterial,
        normalSpecified: normalSpecified,
        normalSmooth: normalSmooth
      },
      indices: {
        aPosition: (0, _parseIndices2.default)(vertexIndices),
        aNormal: (0, _parseIndices2.default)(normalIndices),
        aTexCoord: (0, _parseIndices2.default)(texCoordIndices)
      }
    };
    polyList.push(geometry);
    // Empty current indices buffer.
    vertexIndices = [];
    normalIndices = [];
    texCoordIndices = [];
  }

  function finalizeGeometry() {
    finalizePolyList();
    // Add geometry to output geometries list.
    if (separate) {
      if (noObject) {
        geometries = geometries.concat(polyList);
      } else {
        geometries[objName] = polyList.length === 1 ? polyList[0] : polyList;
      }
    } else {
      geometries = polyList[0];
    }
    polyList = [];
    normalSpecified = false;
    normalSmooth = false;
  }

  function calculateGeometry(geometry) {
    geometry.attributes = attributes;
    if (!geometry.metadata.normalSpecified) {
      if (geometry.metadata.normalSmooth) {
        geometry = (0, _calcSmoothNormals2.default)(geometry);
      } else {
        geometry = (0, _calcNormals2.default)(geometry);
      }
    }
    geometry = (0, _calcTangents2.default)(geometry);
    return geometry;
  }

  // Parser logic starts here.
  var lines = data.split('\n');
  for (var i = 0; i < lines.length; ++i) {
    var line = lines[i].trim();
    // If the line starts with #, this line is a comment - skip the line.
    if (line[0] === '#') continue;
    // Parse the command and the rest.

    var _line$split = line.split(' '),
        command = _line$split[0],
        args = _line$split.slice(1);

    switch (command) {
      // call and csh are not implemented because of security problems
      // (it's not possible to implement it in JS anyway)
      case 'v':
        {
          // Vertex coordinates.
          // weight is not required, so it won't be implemented.
          var _args$map = args.map(parseFloat),
              x = _args$map[0],
              y = _args$map[1],
              z = _args$map[2];

          vertices.push(x, y, z);
          break;
        }
      case 'vn':
        {
          // Normal vector.
          var _args$map2 = args.map(parseFloat),
              _i = _args$map2[0],
              j = _args$map2[1],
              k = _args$map2[2];

          normals.push(_i, j, k);
          normalSpecified = true;
          break;
        }
      case 'vt':
        {
          // Texture coordinates.
          // W is not required since WebGL only supports 2D textures.
          var _args$map3 = args.map(parseFloat),
              u = _args$map3[0],
              v = _args$map3[1];

          if (v == null) v = 0.0;
          texCoords.push(u, v);
          break;
        }
      case 'p':
        {
          // Specifies a point element, however, it is not supported.
          break;
        }
      case 'l':
        {
          // Specifies a line element. not supported.
          break;
        }
      case 'f':
        {
          // Face element. Since arbitary amount of polygons is possible,
          // We have to triangulate the polygon if more than 3 vertices are given.
          var points = args.map(function (arg) {
            var _arg$split = arg.split('/'),
                vertex = _arg$split[0],
                texCoord = _arg$split[1],
                normal = _arg$split[2];

            return {
              vertex: vertex - 1,
              texCoord: texCoord == null ? 0 : texCoord - 1,
              normal: normal == null ? 0 : normal - 1
            };
          });
          for (var _i2 = 1; _i2 < points.length - 1; ++_i2) {
            addFaceIndex(points[0]);
            addFaceIndex(points[_i2]);
            addFaceIndex(points[_i2 + 1]);
          }
          break;
        }
      case 'o':
        {
          // Finalize current buffer if separation is enabled and buffer is not
          // empty.
          if (separate && vertexIndices.length > 0) {
            finalizeGeometry();
          }
          // User defined object name
          objName = args.join(' ');
          break;
        }
      case 'usemtl':
        {
          if (separate && vertexIndices.length > 0) {
            finalizePolyList();
          }
          // Material name.
          objMaterial = args.join(' ');
          break;
        }
      case 'g':
        {
          // Group name, however webglue doesn't support it.
          break;
        }
      case 's':
        {
          if (args[0] === 'off' || args[0] === '0') {
            normalSmooth = false;
          } else {
            normalSmooth = true;
          }
          break;
        }
    }
  }
  // End of file - Build geometry object.
  if (vertexIndices.length > 0) finalizeGeometry();
  // Then, process all geometries - Calculate normals and tangents.
  attributes = {
    aPosition: { data: new Float32Array(vertices), axis: 3 },
    aNormal: { data: new Float32Array(normals), axis: 3 },
    aTexCoord: { data: new Float32Array(texCoords), axis: 2 }
  };

  if (separate) {
    if (noObject) {
      geometries = geometries.map(function (v) {
        return calculateGeometry(v);
      });
    } else {
      for (var key in geometries) {
        if (Array.isArray(geometries[key])) {
          geometries[key] = geometries[key].map(function (v) {
            return calculateGeometry(v);
          });
        } else {
          geometries[key] = calculateGeometry(geometries[key]);
        }
      }
    }
  } else {
    geometries = calculateGeometry(geometries);
  }

  // All done! return the geometry object.
  return geometries;
}