'use strict';

exports.__esModule = true;

var _shaderManager = require('./shaderManager');

var _shaderManager2 = _interopRequireDefault(_shaderManager);

var _geometryManager = require('./geometryManager');

var _geometryManager2 = _interopRequireDefault(_geometryManager);

var _textureManager = require('./textureManager');

var _textureManager2 = _interopRequireDefault(_textureManager);

var _framebufferManager = require('./framebufferManager');

var _framebufferManager2 = _interopRequireDefault(_framebufferManager);

var _stateManager = require('./stateManager');

var _stateManager2 = _interopRequireDefault(_stateManager);

var _renderNode = require('../util/renderNode');

var _renderNode2 = _interopRequireDefault(_renderNode);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var Renderer = function () {
  function Renderer(gl) {
    _classCallCheck(this, Renderer);

    this.gl = gl;

    this.shaders = new _shaderManager2.default(this);
    this.geometries = new _geometryManager2.default(this);
    this.textures = new _textureManager2.default(this);
    this.framebuffers = new _framebufferManager2.default(this);
    this.state = new _stateManager2.default(this);
    // This should be preconfigured in order to set 'standard' attribute
    // indices.
    // Geometries using non-standard attributes cannot be properly cached
    // using VAO.
    // WebGLue geometries use this attributes, however, users can change this
    // to other one.
    // TODO Bind multiple attributes to single index??
    this.attributes = ['aPosition', 'aNormal', 'aTexCoord', 'aTangent'];
    this.width = null;
    this.height = null;
    this.reset();
  }

  Renderer.prototype.reset = function reset() {
    // Should be called after WebGL context reset.
    this.shaders.reset();
    this.geometries.reset();
    this.textures.reset();
    this.framebuffers.reset();
    this.state.reset();
    // Use WebGL extension, if possible.
    this.vao = this.gl.getExtension('OES_vertex_array_object');
    this.uint = this.gl.getExtension('OES_element_index_uint');
    this.instanced = this.gl.getExtension('ANGLE_instanced_arrays');
    this.derivatives = this.gl.getExtension('OES_standard_derivatives');
    this.anisotropic = this.gl.getExtension('EXT_texture_filter_anisotropic');

    this.width = this.gl.drawingBufferWidth;
    this.height = this.gl.drawingBufferHeight;
  };

  Renderer.prototype.render = function render(data) {
    var _this = this;

    if (this.gl.isContextLost()) return false;
    this.state.currentNode = null;
    if (!Array.isArray(data)) return this.renderPass(data, null);
    // Render each pass
    data.forEach(function (pass) {
      return _this.renderPass(pass, null);
    });
  };

  Renderer.prototype.setViewport = function setViewport() {
    var gl = this.gl;
    // TODO Check options and if viewport doesn't exist, continue to use
    // framebuffer's size.
    if (this.framebuffers.current == null) {
      this.width = gl.drawingBufferWidth;
      this.height = gl.drawingBufferHeight;
      gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);
    } else {
      var framebuffer = this.framebuffers.current;
      this.width = framebuffer.width;
      this.height = framebuffer.height;
      gl.viewport(0, 0, framebuffer.width, framebuffer.height);
    }
  };

  Renderer.prototype.renderPass = function renderPass(pass, parent) {
    var _this2 = this;

    if (pass == null || pass === false) return;
    if (Array.isArray(pass)) {
      pass.forEach(function (data) {
        return _this2.renderPass(data, parent);
      });
      return;
    }
    if (typeof pass === 'function') {
      return this.renderPass(pass(parent), parent);
    }
    // Build node
    var node = new _renderNode2.default(pass, parent);
    // Set framebuffers.
    if (pass.framebuffer != null || parent == null) {
      this.framebuffers.use(pass.framebuffer);
      if (node.getOption('viewport') == null) this.setViewport();
    }
    if (pass.options != null) this.state.clear(pass.options);
    // We do 'lazy evaluation' or something like that - to prevent useless
    // WebGL calls.
    if (pass.passes == null) {
      this.shaders.handler = node.shaderHandler;
      this.textures.handler = node.textureHandler;
      // Check frustum culling
      if (!node.shader.frustumCull || this.shaders.checkFrustum(this, node)) {
        this.state.setNode(node);
        if (this.shaders.useNode(node) !== false) {
          this.geometries.use(node.geometry);
          this.geometries.draw();
        }
      }
      // Check mipmap
      if (node.framebuffer != null && node.getOption('mipmap') === true) {
        // TODO This will be a problem if color texture is not specified
        var texture = this.framebuffers.current.options.color;
        if (texture.renderer == null) texture = texture.texture;
        texture.generateMipmap();
      }
    } else {
      if (Array.isArray(pass.passes)) {
        pass.passes.forEach(function (data) {
          return _this2.renderPass(data, node);
        });
      } else {
        this.renderPass(pass.passes, node);
      }
    }
    if (pass.framebuffer !== undefined && parent != null) {
      this.framebuffers.use(parent.framebuffer);
      if (parent.getOption('viewport') == null) this.setViewport();
    }
  };

  return Renderer;
}();

exports.default = Renderer;