'use strict';

exports.__esModule = true;

var _shader2 = require('./shader');

var _shader3 = _interopRequireDefault(_shader2);

var _attachAppendage = require('../util/attachAppendage');

var _attachAppendage2 = _interopRequireDefault(_attachAppendage);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var HEADER_PATTERN = /^\s*#pragma webglue: ([a-z]+)\((([^\),]+)(,\s*(?:[^\),]+))*)\)\s*$/gm;

function parseMetadata(code, governors) {
  var output = {
    features: {},
    counts: {}
  };
  var match = void 0;
  HEADER_PATTERN.lastIndex = 0;
  while ((match = HEADER_PATTERN.exec(code)) != null) {
    var args = match[2].split(/,\s*/g);
    switch (match[1]) {
      case 'feature':
        if (output.features[args[1]] == null) output.features[args[1]] = [];
        output.features[args[1]].push(args[0]);
        break;
      case 'count':
        if (output.counts[args[1]] == null) {
          output.counts[args[1]] = {
            governor: governors[args[2]] || governors.equal,
            defines: []
          };
        }
        output.counts[args[1]].defines.push(args[0]);
        break;
      // TODO other stuff
      default:
        throw new Error('Unknown webglue pragma ' + match[1]);
    }
  }
  return output;
}

function mergeMetadata(vert, frag) {
  var output = {};
  // Categories are same for both vertex and fragment shader
  for (var category in vert) {
    var categoryData = output[category] = {};
    var vertData = vert[category];
    var fragData = frag[category];
    for (var key in vertData) {
      categoryData[key] = { vert: vertData[key] };
    }
    for (var _key in fragData) {
      if (categoryData[_key] == null) categoryData[_key] = {};
      categoryData[_key].frag = fragData[_key];
    }
  }
  return output;
}

function flattenMetadata(metadata) {
  var output = {};

  var _loop = function _loop(category) {
    output[category] = Object.keys(metadata[category]).map(function (key) {
      return Object.assign({}, metadata[category][key], { key: key });
    });
  };

  for (var category in metadata) {
    _loop(category);
  }
  return output;
}

var PreprocessShader = function () {
  function PreprocessShader(renderer, vert, frag) {
    var _this = this;

    _classCallCheck(this, PreprocessShader);

    this.renderer = renderer;
    this.source = { vert: vert, frag: frag };

    var governors = this.renderer.shaders.governors;

    this.metadata = flattenMetadata(mergeMetadata(parseMetadata(vert, governors), parseMetadata(frag, governors)));
    this.useFeatures = this.metadata.features.length !== 0;
    this.useCounts = this.metadata.counts.length !== 0;
    this.invalidators = [];
    if (this.useFeatures) {
      this.metadata.features.forEach(function (v) {
        _this.invalidators.push(v.key);
      });
    }
    if (this.useCounts) {
      this.metadata.counts.forEach(function (v) {
        _this.invalidators.push(v.key);
      });
    }
    this.dirty = true;
    this.currentShader = null;
    this.currentUniforms = {};
    this.currentNode = null;
    // Shaders tree structure: { [flags]: [{counts: ..., shader}, ...], ...}
    // What if there is no 'feature'? It'll fallback to array.
    // What if there is no 'count'? It'll fallback to raw shader object.
    this.shaders = null;
    // ...
    this.frustumCull = vert.indexOf('#pragma webglue: frustumCull\n') !== -1;
  }

  PreprocessShader.prototype.ascendNode = function ascendNode(node) {
    var _this2 = this;

    var uniforms = node.data.uniforms;
    if (uniforms == null) return;
    this.invalidators.forEach(function (key) {
      if (uniforms[key] !== undefined) {
        if (node.parent == null) {
          _this2.currentUniforms[key] = null;
        } else {
          _this2.currentUniforms[key] = node.parent.getUniform(key);
        }
        _this2.dirty = true;
      }
    });
  };

  PreprocessShader.prototype.descendNode = function descendNode(node) {
    var _this3 = this;

    var uniforms = node.data.uniforms;
    if (uniforms == null) return;
    this.invalidators.forEach(function (key) {
      if (uniforms[key] !== undefined) {
        // TODO Check if count/feature is still valid
        _this3.currentUniforms[key] = uniforms[key];
        _this3.dirty = true;
      }
    });
  };

  PreprocessShader.prototype.getShader = function getShader() {
    var _this4 = this;

    var uniforms = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : this.currentUniforms;

    if (uniforms === this.currentUniforms && !this.dirty) {
      return this.currentShader;
    }
    this.dirty = false;
    if (this.useFeatures) {
      if (this.shaders == null) this.shaders = {};
      var vertDefines = [];
      var fragDefines = [];
      // Build current feature map
      var featureKey = this.metadata.features.reduce(function (prev, v) {
        if (uniforms[v.key]) {
          vertDefines.push.apply(vertDefines, v.vert || []);
          fragDefines.push.apply(fragDefines, v.frag || []);
          return (prev << 1) + 1;
        } else {
          return prev << 1;
        }
      }, 0);
      if (this.useCounts) {
        var selected = void 0;
        if (this.shaders[featureKey] == null) {
          selected = this.shaders[featureKey] = [];
        } else {
          selected = this.shaders[featureKey];
        }
        // Scan the counts list and find good match
        // TODO This is too costy and bad
        var match = selected.find(function (entry) {
          return _this4.metadata.counts.every(function (v) {
            var governor = v.vert && v.vert.governor || v.frag && v.frag.governor;
            return governor.checker(entry.uniforms[v.key], uniforms[v.key]);
          });
        });
        if (match == null) {
          var uniformData = {};
          this.metadata.counts.forEach(function (v) {
            var governor = v.vert && v.vert.governor || v.frag && v.frag.governor;
            var value = governor.allocator(uniforms[v.key]);
            vertDefines.push.apply(vertDefines, (v.vert && v.vert.defines || []).map(function (field) {
              return field + ' ' + value;
            }));
            fragDefines.push.apply(fragDefines, (v.frag && v.frag.defines || []).map(function (field) {
              return field + ' ' + value;
            }));
            uniformData[v.key] = value;
          });
          // Create the shader
          var vertStr = vertDefines.map(function (v) {
            return '#define ' + v + '\n';
          }).join('');
          var fragStr = fragDefines.map(function (v) {
            return '#define ' + v + '\n';
          }).join('');
          var shader = new _shader3.default(this.renderer, (0, _attachAppendage2.default)(this.source.vert, vertStr), (0, _attachAppendage2.default)(this.source.frag, fragStr));
          shader.parent = this;
          selected.push({ shader: shader, uniforms: uniformData });
          this.currentShader = shader;
          return shader;
        } else {
          this.currentShader = match.shader;
          return match.shader;
        }
      } else {
        if (this.shaders[featureKey] == null) {
          // Create the shader
          var _vertStr = vertDefines.map(function (v) {
            return '#define ' + v + '\n';
          }).join('');
          var _fragStr = fragDefines.map(function (v) {
            return '#define ' + v + '\n';
          }).join('');
          var _shader = new _shader3.default(this.renderer, (0, _attachAppendage2.default)(this.source.vert, _vertStr), (0, _attachAppendage2.default)(this.source.frag, _fragStr));
          _shader.parent = this;
          this.shaders[featureKey] = _shader;
        }
        var _selected = this.shaders[featureKey];
        this.currentShader = _selected;
        return _selected;
      }
    } else {
      if (this.shaders == null) {
        this.shaders = new _shader3.default(this.renderer, this.source.vert, this.source.frag);
        this.shaders.parent = this;
      }
      this.currentShader = this.shaders;
      return this.shaders;
    }
  };

  PreprocessShader.prototype.use = function use(uniforms, current) {
    var shader = this.getShader(uniforms);
    return shader.use(uniforms, current);
  };

  PreprocessShader.prototype.dispose = function dispose() {
    // Dispose all shader objects. This is awkward since we have to write
    // the same code twice.
    if (this.useFeatures) {
      for (var key in this.shaders) {
        if (this.useCounts) {
          this.shaders[key].forEach(function (v) {
            return v.shader.dispose();
          });
        } else {
          this.shaders[key].dispose();
        }
      }
    } else {
      if (this.useCounts) {
        this.shaders.forEach(function (v) {
          return v.shader.dispose();
        });
      } else {
        if (this.shaders != null) this.shaders.dispose();
      }
    }
  };

  return PreprocessShader;
}();

exports.default = PreprocessShader;