'use strict';

exports.__esModule = true;

var _parseUniform = require('../util/parseUniform');

var _parseUniform2 = _interopRequireDefault(_parseUniform);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function compileShader(gl, data, type) {
  var shader = gl.createShader(type);
  gl.shaderSource(shader, data);
  gl.compileShader(shader);
  if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS) && !gl.isContextLost()) {
    throw new Error('Shader compilation failed: ' + gl.getShaderInfoLog(shader));
  }
  return shader;
}

var Shader = function () {
  function Shader(renderer, vert, frag) {
    _classCallCheck(this, Shader);

    this.renderer = renderer;
    this.source = { vert: vert, frag: frag };
    // TODO This should allow dynamic pipeline.. or something.
    // This only does simple AABB though.
    this.frustumCull = vert.indexOf('#pragma webglue: frustumCull\n') !== -1;
    this.program = null;
    this.attributes = {};
    this.uniforms = {};
    this.uniformTypes = {};
    this.currentNode = null;
  }

  Shader.prototype.upload = function upload() {
    if (this.program !== null) return;
    var gl = this.renderer.gl;
    // Compile shaders
    this.shader = {
      vert: compileShader(gl, this.source.vert, gl.VERTEX_SHADER),
      frag: compileShader(gl, this.source.frag, gl.FRAGMENT_SHADER)
    };
    // Link shaders
    var program = gl.createProgram();
    // Use standard geometries if possible
    this.renderer.attributes.forEach(function (name, index) {
      gl.bindAttribLocation(program, index, name);
    });
    gl.attachShader(program, this.shader.vert);
    gl.attachShader(program, this.shader.frag);
    gl.linkProgram(program);

    if (!gl.getProgramParameter(program, gl.LINK_STATUS) && !gl.isContextLost()) {
      throw new Error('Shader program linking failed: ' + gl.getProgramInfoLog(program));
    }

    // Clean up the shaders
    gl.detachShader(program, this.shader.vert);
    gl.detachShader(program, this.shader.frag);
    gl.deleteShader(this.shader.vert);
    gl.deleteShader(this.shader.frag);

    this.program = program;
    this.attributes = {};
    this.uniforms = {};

    // Load program information
    var attributeSize = gl.getProgramParameter(program, gl.ACTIVE_ATTRIBUTES);
    for (var i = 0; i < attributeSize; ++i) {
      var attribute = gl.getActiveAttrib(program, i);
      var name = attribute.name;
      this.attributes[name] = gl.getAttribLocation(program, name);
    }

    var textureId = 0;

    // Since uniform supports struct and array, this gets pretty tricky..
    var uniformSize = gl.getProgramParameter(program, gl.ACTIVE_UNIFORMS);
    for (var _i = 0; _i < uniformSize; ++_i) {
      var uniform = gl.getActiveUniform(program, _i);
      var _name = uniform.name,
          typeId = uniform.type,
          size = uniform.size;

      if (size > 1) {
        // Array has been received - We have to process each uniform
        // manually, since uniform list only has one uniform representing
        // index 0.
        for (var _i2 = 0; _i2 < size; ++_i2) {
          // Something doesn't feel right about this. It works but it's weird.
          var newName = _name.replace('[0]', '[' + _i2 + ']');
          var location = gl.getUniformLocation(program, newName);
          var texturePos = null;
          if (typeId === gl.SAMPLER_2D || typeId === gl.SAMPLER_CUBE) {
            texturePos = textureId++;
          }
          this._addUniform(newName, {
            name: newName,
            type: typeId,
            location: location,
            texture: texturePos
          });
        }
      } else {
        var _location = gl.getUniformLocation(program, _name);
        var _texturePos = null;
        if (typeId === gl.SAMPLER_2D || typeId === gl.SAMPLER_CUBE) {
          _texturePos = textureId++;
        }
        this._addUniform(_name, {
          name: _name,
          type: typeId,
          location: _location,
          texture: _texturePos
        });
      }
    }
  };

  Shader.prototype._addUniform = function _addUniform(name, metadata) {
    // Support raw string (.[] is reserved anyway)
    this.uniforms[name] = metadata;
    var parent = this.uniforms;
    var parentArray = false;
    var parentIndex = null;
    var truncName = name;
    var hasFinished = false;
    do {
      // Check for array..
      var arrayStartLoc = truncName.indexOf('[');
      if (arrayStartLoc !== -1) {
        var arrayEndLoc = truncName.indexOf(']');
        var fieldName = truncName.slice(0, arrayStartLoc);
        var index = parseInt(truncName.slice(arrayStartLoc + 1, arrayEndLoc));
        truncName = truncName.slice(arrayEndLoc + 1);

        var parentField = parentArray ? parentIndex : fieldName;
        if (parent[parentField] == null) {
          parent[parentField] = [];
        }
        parent = parent[parentField];

        parentArray = true;
        parentIndex = index;
      }
      var objectStartLoc = truncName.indexOf('.');
      if (objectStartLoc !== -1) {
        var _fieldName = truncName.slice(0, objectStartLoc);
        truncName = truncName.slice(objectStartLoc + 1);

        var _parentField = parentArray ? parentIndex : _fieldName;
        if (parent[_parentField] == null) {
          parent[_parentField] = {};
        }
        parent = parent[_parentField];

        parentArray = false;
      } else {
        var _fieldName2 = truncName;

        var _parentField2 = parentArray ? parentIndex : _fieldName2;
        if (parent[_parentField2] == null) {
          parent[_parentField2] = {};
        }
        parent[_parentField2] = metadata;

        hasFinished = true;
      }
    } while (!hasFinished);
  };

  Shader.prototype.getShader = function getShader() {
    return this;
  };

  Shader.prototype.use = function use(uniforms, current) {
    var gl = this.renderer.gl;
    if (this.program === null) this.upload();
    if (current !== this) gl.useProgram(this.program);
    if (uniforms != null) this.setUniforms(uniforms);
    return this;
  };

  Shader.prototype.setUniforms = function setUniforms(originalValues) {
    var uniforms = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this.uniforms;

    var values = originalValues;
    if (values === undefined) return;
    if (typeof originalValues === 'function') {
      values = originalValues(this, this.renderer);
    }
    // TODO We need to make sure that this is indeed a 'leaf' node,
    // otherwise we won't be able to use 'type' as a variable name
    if (typeof uniforms.type === 'number') {
      return this.setUniform(values, uniforms);
    }
    // Disable setting to 0 because it's meaningless.
    if (values === false) return;
    // Set all children to 0...
    /*
    if (values === false) {
      if (Array.isArray(uniforms)) {
        for (let i = 0; i < uniforms.length; ++i) {
          this.setUniforms(false, uniforms[i]);
        }
      } else {
        for (let i in uniforms) {
          this.setUniforms(false, uniforms[i]);
        }
      }
    }
    */
    if (Array.isArray(values)) {
      for (var i = 0; i < values.length; ++i) {
        if (uniforms[i] == null) continue;
        this.setUniforms(values[i], uniforms[i]);
      }
    } else {
      for (var _i3 in values) {
        if (uniforms[_i3] == null) continue;
        this.setUniforms(values[_i3], uniforms[_i3]);
      }
    }
  };

  Shader.prototype.setUniform = function setUniform(val, metadata) {
    var gl = this.renderer.gl;
    var key = metadata.location;
    if (key == null) return;
    var value = (0, _parseUniform2.default)(gl, val, metadata.type);
    switch (metadata.type) {
      case gl.FLOAT_VEC2:
        gl.uniform2fv(key, value);
        break;
      case gl.FLOAT_VEC3:
        gl.uniform3fv(key, value);
        break;
      case gl.FLOAT_VEC4:
        gl.uniform4fv(key, value);
        break;
      case gl.INT_VEC2:
      case gl.BOOL_VEC2:
        gl.uniform2iv(key, value);
        break;
      case gl.INT_VEC3:
      case gl.BOOL_VEC3:
        gl.uniform3iv(key, value);
        break;
      case gl.INT_VEC4:
      case gl.BOOL_VEC4:
        gl.uniform4iv(key, value);
        break;
      case gl.BOOL:
      case gl.BYTE:
      case gl.UNSIGNED_BYTE:
      case gl.SHORT:
      case gl.UNSIGNED_SHORT:
      case gl.INT:
      case gl.UNSIGNED_INT:
        gl.uniform1i(key, value);
        break;
      case gl.FLOAT:
        gl.uniform1f(key, value);
        break;
      case gl.FLOAT_MAT2:
        gl.uniformMatrix2fv(key, false, value);
        break;
      case gl.FLOAT_MAT3:
        gl.uniformMatrix3fv(key, false, value);
        break;
      case gl.FLOAT_MAT4:
        gl.uniformMatrix4fv(key, false, value);
        break;
      case gl.SAMPLER_2D:
      case gl.SAMPLER_CUBE:
        // Call texture handler (if any)
        if (this.renderer.textures.handler != null) {
          value = this.renderer.textures.handler(value);
        }
        gl.uniform1i(key, this.renderer.textures.use(metadata.texture, value));
        break;
    }
  };

  Shader.prototype.dispose = function dispose() {
    // Nothing to do yet
    var gl = this.renderer.gl;
    gl.deleteProgram(this.program);
    this.program = null;
  };

  return Shader;
}();

exports.default = Shader;