'use strict';

exports.__esModule = true;
exports.isSource = isSource;
exports.isLoaded = isLoaded;

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var OPTIONS_KEY = {
  magFilter: 0x2800,
  minFilter: 0x2801,
  wrapS: 0x2802,
  wrapT: 0x2803,
  maxAnisotropy: 0x84FE
};

function isSource(source) {
  if (source instanceof HTMLElement) return true;
  if (source instanceof ImageData) return true;
  return false;
}

function isPromise(promise) {
  if (promise == null) return false;
  return typeof promise.then === 'function';
}

function setUpPromise(texture, source, position) {
  source.then(function (value) {
    if (position != null) texture.options.source[position] = value;else texture.options.source = value;
  });
}

function isLoaded(source) {
  if (isPromise(source)) return false;
  if (isSource(source)) {
    // Check readystate...
    if (source.readyState != null && source.readyState !== 4) {
      return false;
    }
    // If there is complete property and it is false, return false.
    if (source.complete === false) return false;
  }
  return true;
}

function getWidth(options, renderer) {
  return renderer.gl.drawingBufferWidth;
}

function getHeight(options, renderer) {
  return renderer.gl.drawingBufferHeight;
}

var Texture = function () {
  function Texture(renderer, options) {
    var _this = this;

    _classCallCheck(this, Texture);

    this.renderer = renderer;
    this.options = options;

    // TODO immutable
    if (Array.isArray(this.options.source)) {
      this.options.source = this.options.source.map(function (source, index) {
        if (typeof source === 'string') {
          var image = new Image();
          image.src = source;
          return image;
        } else if (isPromise(source)) {
          setUpPromise(_this, source, index);
        }
        return source;
      });
    } else {
      if (typeof this.options.source === 'string') {
        var image = new Image();
        image.src = this.options.source;
        this.options.source = image;
      } else if (isPromise(this.options.source)) {
        setUpPromise(this, this.options.source);
      }
    }

    this.unit = -1;
    this.texture = null;
    this.loaded = false;
    // Use this to reupload only the texture
    this.valid = false;

    this.mipmapPending = false;
    this.varyingSize = false;

    this.width = null;
    this.height = null;
  }

  Texture.prototype.init = function init() {
    var gl = this.renderer.gl;
    this.texture = gl.createTexture();
  };

  Texture.prototype.uploadTexture = function uploadTexture(target, source) {
    var gl = this.renderer.gl;
    this.varyingSize = false;
    var format = this.options.format;
    if (typeof format === 'function') {
      format = format(this.options, this.renderer);
    }
    var type = this.options.type;
    if (typeof type === 'function') type = type(this.options, this.renderer);
    if (isSource(source)) {
      gl.texImage2D(target, 0, format, format, type, source);

      if (this.width !== null && source.width !== this.width && this.height !== null && source.height !== this.height) {
        throw new Error('Size mismatch');
      }
      this.width = source.width;
      this.height = source.height;
    } else {
      // width and height shouldn't be 0, right?
      var width = this.options.width || getWidth;
      var height = this.options.height || getHeight;
      if (typeof width === 'function') {
        width = width(this.options, this.renderer);
        this.varyingSize = true;
      }
      if (typeof height === 'function') {
        height = height(this.options, this.renderer);
        this.varyingSize = true;
      }
      gl.texImage2D(target, 0, format, width, height, 0, format, type, source);

      this.width = width;
      this.height = height;
    }
  };

  Texture.prototype.validateSize = function validateSize() {
    var width = this.options.width || getWidth;
    var height = this.options.height || getHeight;
    if (typeof width === 'function') {
      width = width(this.options, this.renderer);
    }
    if (typeof height === 'function') {
      height = height(this.options, this.renderer);
    }
    return this.width === width && this.height === height;
  };

  Texture.prototype.reupload = function reupload() {
    var gl = this.renderer.gl;
    var source = this.options.source;
    // Make sure everything is loaded and load...
    if (Array.isArray(source)) {
      if (!source.every(isLoaded)) return false;
      gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, this.options.params.flipY);
      for (var i = 0; i < 6; ++i) {
        this.uploadTexture(gl.TEXTURE_CUBE_MAP_POSITIVE_X + i, source[i]);
      }
    } else {
      if (!isLoaded(source)) return false;
      gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, this.options.params.flipY);
      this.uploadTexture(gl.TEXTURE_2D, source);
    }
    this.valid = true;
    return true;
  };

  Texture.prototype.upload = function upload() {
    var gl = this.renderer.gl;
    if (this.texture == null) this.init();
    var target = void 0;
    var source = this.options.source;
    if (Array.isArray(source)) {
      target = gl.TEXTURE_CUBE_MAP;
    } else {
      target = gl.TEXTURE_2D;
    }
    this.target = target;
    gl.bindTexture(target, this.texture);

    this.width = null;
    this.height = null;

    if (!this.reupload()) return false;

    // Set texture parameters
    for (var key in this.options.params) {
      var value = this.options.params[key];
      if (typeof value === 'function') {
        value = value(this.options, this.renderer);
      }
      if (key === 'flipY') continue;
      if (key === 'mipmap') {
        if (source == null) continue;
        if (value) gl.generateMipmap(target);
        continue;
      }
      if (key === 'maxAnisotropy') {
        if (this.renderer.anisotropic) {
          gl.texParameterf(target, OPTIONS_KEY[key], value);
        }
        continue;
      }
      gl.texParameteri(target, OPTIONS_KEY[key], value);
    }
    this.loaded = true;
    this.valid = true;
    // All done!
  };

  Texture.prototype.generateMipmap = function generateMipmap() {
    this.mipmapPending = true;
  };

  Texture.prototype.use = function use(unit, isFramebuffer) {
    var gl = this.renderer.gl;
    this.unit = unit;
    gl.activeTexture(gl.TEXTURE0 + unit);
    // Reupload the texture if required.
    if (this.texture == null) this.init();
    if (!this.loaded || this.varyingSize && !this.validateSize()) {
      this.upload();
    } else {
      gl.bindTexture(this.target, this.texture);
      if (!this.valid) this.reupload();
      if (this.mipmapPending && !isFramebuffer) {
        this.mipmapPending = false;
        gl.generateMipmap(this.target);
      }
    }
  };

  Texture.prototype.dispose = function dispose() {
    var gl = this.renderer.gl;
    gl.deleteTexture(this.texture);
    this.texture = null;
    this.loaded = false;
  };

  return Texture;
}();

exports.default = Texture;